#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef PATH_MAX
#define PATH_MAX 2048
#endif

#define def_bitrate 128000
#define def_cut_start 0  // cut from
#define def_cut_end 100   // cut to

// Search for next occurence of 0xFFFB in 'file' at or after 'start'
static long find_mp3_header(FILE *file, long start)
{
	unsigned char ch;
//	unsigned char match[2] = {0xFF, 0xFB};
	int n = 0;
	
	if (fseek(file, start, SEEK_SET) < 0) {
		fprintf(stderr, "Failed to seek to %lu!\n", start);
		return -1;
	}
	do {
		if (fread(&ch, 1, 1, file) < 1) break;
		if (n == 0) {
			if (ch == 0xFF) n++;
			else n = 0;
		} else {
			if ((ch == 0xFB) || (ch == 0xF3)) n++;
			else if (ch == 0xFF) n = 1;
			else n = 0;
		}
//		if (ch == match[n]) n++;          // match next?
//		else if (ch == match[0]) n = 1;  // match first?
//		else n = 0;                     // no match
//		if (n == sizeof(match)) 
		if (n > 1) 
			return ftell(file) - n;
	} while (1);
	return -1;
}

// Copy 'size' bytes from open file 'in' to open file 'out'
// returns number of bytes copied, or -1
static long fcopy(FILE *in, FILE *out, long size)
{
	unsigned char buffer[64 * 1024];
	long block, chunk, offs, copied = 0;
	
	block = sizeof(buffer);
	do {
		offs = 0;
		if (block > (size - copied)) block = size - copied;
		do {
			chunk = fread(&buffer[offs], 1, block - offs, in);
			if (chunk <= 0) break;
			offs += chunk;
		} while (offs < block);
		if (offs < block) block = offs;
		offs = 0;
		do {
			chunk = fwrite(&buffer[offs], 1, block - offs, out);
			if (chunk <= 0) break;
			offs += chunk;
		} while (offs < block);
		copied += offs;
		if (offs < sizeof(buffer)) break;
	} while (copied < size);
	return copied;
}

int main(int argc, char *argv[])
{
	int arg_in, arg_out, c;
	extern char *optarg;
	extern int optind, opterr, optopt;
	char *endptr;
	int param;
	long long bitrate, cut_start, cut_end;
	FILE *in, *out;
	long long start, end, size, copied;
	char target[PATH_MAX];
	
	// Get parameter
	bitrate = def_bitrate;
	cut_start = def_cut_start;
	cut_end = def_cut_end;
	while (1) {
		c = getopt(argc, argv, "b:s:e:");
		if (c < 0) break;
		switch (c) {
			case 'b':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') bitrate = param * 1000;
				break;
			case 's':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') cut_start = param;
				break;
			case 'e':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') cut_end = param;
				break;
			default:
				optind = argc;
		}
	}
	if (optind < argc) arg_in = optind++;
	else {
		printf("Removes a section from an MPEG1 Layer III audio file. (c) 2003 Christian Wolff.\n");
		printf("usage: %s <input file> [<output file>] [-b <bitrate> (128 kBit/s)] [-s <cut start> (0/10 sec.)] [-e <cut end> (100/10 sec.)]\n", argv[0]);
		exit(0);
	}
	if (optind < argc) arg_out = optind++;
	else arg_out = 0;
	
	// Print status
	printf("Source file: %s\n", argv[arg_in]);
	if (arg_out) printf("Target file: %s\n", argv[arg_out]);
	printf("Bitrate: %d kBit/sec., cut out from %d.%1d to %d.%1d sec.\n", 
		bitrate / 1000, cut_start / 10, cut_start % 10, (cut_start + cut_end) / 10, (cut_start + cut_end) % 10);
	
	// Open source file
	in = fopen(argv[arg_in], "rb");
	if (in == NULL) {
		fprintf(stderr, "File not found: %s\n", argv[arg_in]);
		exit(1);
	}
	
	// Find start and end, aligned to mp3 block
	start = find_mp3_header(in, cut_start * (bitrate >> 3) / 10);
	if (start < 0) {
		fprintf(stderr, "Failed to find MP3 header %d.%1d seconds into the file\n", cut_start / 10, cut_start % 10);
		exit(1);
	}
	end = find_mp3_header(in, start + cut_end * (bitrate >> 3) / 10);
	if (end < 0) {
		fprintf(stderr, "Failed to find MP3 header %d.%1d seconds into the file, deleting until end of file\n", (cut_start + cut_end) / 10, (cut_start + cut_end) % 10);
		fseek(in, 0, SEEK_END);
		end = ftell(in);
	}
	printf("Cutting from 0x%X to 0x%X,\n", start, end);
	
	// Open destination file
	if (arg_out) {
		if (! strncmp(argv[arg_out], argv[arg_in], PATH_MAX)) {
			fprintf(stderr, "Input and Output can't be the same file!\n");
			exit(1);
		}
		printf("Writing to %s\n", argv[arg_out]);
		out = fopen(argv[arg_out], "wb");
		if (out == NULL) fprintf(stderr, "Error writing file: %s\n", argv[arg_out]);
	} else {
		strncpy(target, argv[arg_in], PATH_MAX);
		strncat(target, "_edit.mp3", PATH_MAX);
		printf("Writing to %s\n", target);
		out = fopen(target, "wb");
		if (out == NULL) fprintf(stderr, "Error writing file: %s\n", target);
	}
	if (out == NULL) exit(1);
	
	// Get source file size
	if (fseek(in, 0, SEEK_END) < 0) {
		fprintf(stderr, "Failed to seek to end of input file!\n");
		exit(1);
	}
	size = ftell(in);
	
	// Copy start of file
	printf("Copying start of file (%ld bytes)\n", start);
	rewind(in);
	copied = fcopy(in, out, start);
	if (copied < start) {
		fprintf(stderr, "Failed to copy %ld bytes, wrote %ld bytes instead!\n", start, copied);
		exit(1);
	}
	
	// Copy end of file
	printf("Copying end of file (%ld - %ld = %ld bytes)\n", size, end, size - end);
	if (fseek(in, end, SEEK_SET) < 0) {
		fprintf(stderr, "Failed to seek to %ld in input file!\n", end);
		exit(1);
	}
	copied = fcopy(in, out, size - end);
	if (copied < size - end) {
		fprintf(stderr, "Failed to copy %ld bytes, wrote %ld bytes instead!\n", size - end, copied);
		exit(1);
	}
	
	// Close files
	fclose(in);
	fclose(out);
	
	return 0;
}

